#
# NYCode.R
# Stella Watson and Andrew Brown
# 23 Dec 2017
#
# This file contains code for the NY election example in the manuscript. The
# data are available at the Harvard Dataverse. See manuscript for details.
rm(list= ls())

#####Load Packages
library(BayesLogit)  # Uses Polya-Gamma for Bayesian logistic regression
library(car)  # "companion to applied regression"
library(coda) 
library(ExtDist)  # More probability distributions than standard 
library(foreign)
library(ggplot2)
library(inline)  # allows inline calls to C, C++, and FORTRAN
library(invgamma)  # Makes available the inverse Gamma dist. directly
library(kriging)  # Kriging algorithm to plot geographical data
library(maps)
library(maptools)
library(MASS)
library(Matrix)
library(matrixcalc)  # additional matrix operations
library(msm)  # package for fitting multi-state Markov models
library(mvtnorm)  # mvn and mvt distributions
library(PBSmapping)
library(SDMTools)  # "Species distribution modeling tools"
library(snow)
library(sp)  # classes and methods for spatial data
library(rARPACK)  # interfaces to spectra package, which solves large-scale
# eigenspace problems
library(Rcpp)
library(rgdal)  # Geospatial data abstraction library. Uses classes created in
# the sp package.
library(rgeos)
library(rgl)  
library(scatterplot3d)
library(sp)
library(spdep)
library(spam)
library(SparseM)  # package for sparse linear algebra. Spam supposedly improves
# on this by storing the sparsity structure.
library(truncnorm)  # Truncated normal distribution

library(profmem)  # facilitates easy memory profiling

###Read in the Files
path = '/Users/stellawatson/Box Sync/Clemson/Research/Chromatic Sampler/arcgis files/ny_shapefile'
ny_shape <- readOGR(dsn = path, layer = "ny_final")

###Hyperparameters
a_tau = 1
b_tau = 1
upsilon.s = 1000

### generate the adjacency matrix
nbhd.list = poly2nb(ny_shape)
W = nb2mat(nbhd.list,style = "B")
W = as.spam(W)
D = diag(colSums(W))
D = as.spam(D)
omega = 0.995
DW = as.spam(D - omega*W)
S = dim(DW)[1]
#A = cbind(A,A)
#A = rbind(A,A)
V = dim(DW)[1]
I <- diag.spam(rep(1,times= V))
Rs <- chol(DW + I)

###Response Data
main.df = data.frame(as.numeric(ny_shape$GEOID10),ny_shape$TOTAL,
                     ny_shape$GOV_DVOTE_,seq(1,S),stringsAsFactors = FALSE)
names(main.df) = c("TractID","TotalVotes","DemVotes","W_ID")
ind = main.df$TotalVotes >0
main.df = main.df[ind,]
n= sum(main.df$TotalVotes)
nct = dim(main.df)[1]

###Create IDS for CGS
# This is the code that finds the graph coloring for the chromatic sampler. The
# greedy algorithm in this case finds a 7-coloring.
id<-1:S
id.grp1<-NULL
id.rem1<-NULL
for(i in 1:S){
  if(sum(id[i]==id.rem1)==0){
    id.grp1<-c(id.grp1,id[i])
    id.rem1<-c(id.rem1,id[as.matrix(W[id[i],])==1])
  }
}
id.grp2<-NULL
id.rem2<-id.grp1
for(i in 1:S){
  if(sum(id[i]==id.rem2)==0){
    id.grp2<-c(id.grp2,id[i])
    id.rem2<-c(id.rem2,id[as.matrix(W[id[i],])==1])
  }
}
id.grp3<-NULL
id.rem3<-c(id.grp1,id.grp2)
for(i in 1:S){
  if(sum(id[i]==id.rem3)==0){
    id.grp3<-c(id.grp3,id[i])
    id.rem3<-c(id.rem3,id[as.matrix(W[id[i],])==1])
  }
}
id.grp4<-NULL
id.rem4<-c(id.grp1,id.grp2,id.grp3)
for(i in 1:S){
  if(sum(id[i]==id.rem4)==0){
    id.grp4<-c(id.grp4,id[i])
    id.rem4<-c(id.rem4,id[as.matrix(W[id[i],])==1])
  }
}
id.grp5<-NULL
id.rem5<-c(id.grp1,id.grp2,id.grp3,id.grp4)
for(i in 1:S){
  if(sum(id[i]==id.rem5)==0){
    id.grp5<-c(id.grp5,id[i])
    id.rem5<-c(id.rem5,id[as.matrix(W[id[i],])==1])
  }
}
id.grp6<-NULL
id.rem6<-c(id.grp1,id.grp2,id.grp3,id.grp4,id.grp5)
for(i in 1:S){
  if(sum(id[i]==id.rem6)==0){
    id.grp6<-c(id.grp6,id[i])
    id.rem6<-c(id.rem6,id[as.matrix(W[id[i],])==1])
  }
}
id.grp7<-NULL
id.rem7<-c(id.grp1,id.grp2,id.grp3,id.grp4,id.grp5,id.grp6)
for(i in 1:S){
  if(sum(id[i]==id.rem7)==0){
    id.grp7<-c(id.grp7,id[i])
    id.rem7<-c(id.rem7,id[as.matrix(W[id[i],])==1])
  }
}

U = matrix(1,nct,1)

##Make G Matrix
r.i = seq(1,nct,by = 1)
xv = rep(1,nct)
ci = rep(-99,nct)
for(j in 1:S){
  ind = which(main.df$W_ID == j)
  if(length(ind) >0){
    ci[ind] = j
  }
  print(j)
}
G.Mat = sparseMatrix(i = r.i, j = ci, x = xv, dims = c(nct,S))
G.Mat.spam = as.spam(as.matrix(G.Mat))
G.ind = is.element(seq(1,S),main.df$W_ID)
t.G <-function(v){
  res = rep(0,S)
  res[main.df$W_ID] = v
  return(res)
}

###Initialization
mi = main.df$TotalVotes
ind = mi==0
mi[ind] = 1
Y = main.df$DemVotes
H = Y - mi/2

####Check
i = 459
main.df[i,]
ci[i]
Y[i]
mi[i]


block.gibbs<-function(G, burn, upsilon.g, xi.g, tau.g, VERBOSE){ 
  
  G.xi.g = xi.g[G.ind]
  omega.g = omega
  xi.array = matrix(99,S,G)
  tau.array = matrix(99,1,G)
  upsilon.array = matrix(99,1,G)
  upsilon.array[1,1] = upsilon.g
  xi.array[,1] = xi.g
  tau.array[1,1] = tau.g
  
  ###Sampling Loop
  start.time = Sys.time()
  for(g in 1:G){
    
    eta.g = as.vector(U*upsilon.g +G.xi.g)
    w.g = rpg(nct,as.numeric(mi),as.numeric(eta.g))
    
    ###Sample Upsilon
    Sig.upsilon = as.numeric(1/(sum(w.g*U^2) + 1/upsilon.s))
    Mn.Upsilon = as.numeric(Sig.upsilon*(-1*sum(w.g*U*G.xi.g) + sum(U*H)))
    upsilon.g = rnorm(1,Mn.Upsilon,sqrt(Sig.upsilon))
    upsilon.array[1,g] = upsilon.g
    
    ###Sample Xi
    Prec.xi =   as.spam(I*t.G(w.g) + DW/tau.g)
    
    
    Prec.Mu.xi = t.G(-1*w.g*U*upsilon.g + H)  
    xi.g = c(rmvnorm.canonical(1, Prec.Mu.xi, Prec.xi, Rstruct= Rs))
    xi.g = xi.g - mean(xi.g)
    G.xi.g = xi.g[G.ind]
    xi.array[,g] = xi.g
    
    ###Sample tau
    tau.g = rinvgamma(1,S/2+a_tau, 1/2*sum(xi.g*((DW)%*%xi.g)) + b_tau) 
    tau.array[1,g] = tau.g
    if(VERBOSE){ 
      print(g)
    }
  }
  
  total.time = Sys.time() - start.time
  return.list = list(ups= upsilon.array, tau= tau.array,
                     xi= xi.array, time= total.time, iters= G, burn.in= burn)
  
}  # End function block.gibbs


chromatic.gibbs<-function(G,burn,upsilon.g,xi.g,tau.g,VERBOSE){ 
  
  G.xi.g = xi.g[G.ind]
  omega.g = omega
  xi.array = matrix(99,S,G)
  tau.array = matrix(99,1,G)
  upsilon.array = matrix(99,1,G)
  upsilon.array[1,1] = upsilon.g
  xi.array[,1] = xi.g
  tau.array[1,1] = tau.g
  
  id.gr <- list(id.grp1, id.grp2, id.grp3, id.grp4, id.grp5, id.grp6, id.grp7)
  
  ###Sampling Loop
  start.time = Sys.time()
  for(g in 1:G){
    
    eta.g = as.vector(U*upsilon.g +G.xi.g)
    w.g = rpg(nct,as.numeric(mi),as.numeric(eta.g))
    
    ###Sample Upsilon
    Sig.upsilon = as.numeric(1/(sum(w.g*U^2) + 1/upsilon.s))
    Mn.Upsilon = as.numeric(Sig.upsilon*(-1*sum(w.g*U*G.xi.g) + sum(U*H)))
    upsilon.g = rnorm(1,Mn.Upsilon,sqrt(Sig.upsilon))
    upsilon.array[1,g] = upsilon.g
    
    ###Sample Xi
    v1 = t.G(w.g)
    sig.xi = 1/(v1 + diag(D)/tau.g)
    v1= v1*upsilon.g
    v2 = t.G(H)
    
    for (i in 1:7){
      mu.xi = sig.xi[id.gr[[i]]]*(omega/tau.g*W[id.gr[[i]],]%*%xi.g - 
                                    v1[id.gr[[i]]] + v2[id.gr[[i]]])
      xi.g[id.gr[[i]]] = rnorm(length(id.gr[[i]]),as.vector(mu.xi),
                               as.vector(sqrt(sig.xi[id.gr[[i]]])))
    }
    xi.g = xi.g - mean(xi.g)
    G.xi.g = xi.g[G.ind]
    xi.array[,g] = xi.g
    
    ###Sample tau
    tau.g = rinvgamma(1,S/2+a_tau, 1/2*sum(xi.g*(DW%*%xi.g)) + b_tau)
    tau.array[1,g] = tau.g
    if(VERBOSE){
      print(g)
    }
  }
  total.time = Sys.time() - start.time
  return.list = list(upsilon.array,tau.array,xi.array,total.time,G,burn)
}



###First Initialization
upsilon.i = 0
xi.i = matrix(0,S,1)
tau.i = 1

###Block
block1 = block.gibbs(10000,5000,upsilon.i,xi.i,tau.i,TRUE)
ups.b1 = block1[[1]]
tau.b1 = block1[[2]]
xi.b1 = block1[[3]]
time.b1 = block1[[4]]
G.b1 = block1[[5]]
burn.b1 = block1[[6]]
ups.hat.b1 = mean(ups.b1[,burn.b1:G.b1])
tau.hat.b1 = mean(tau.b1[,burn.b1:G.b1])
xi.hat.b1 = apply(xi.b1[,burn.b1:G.b1],1,mean)  

###Chromatic
chromatic1 = chromatic.gibbs(10000,5000,upsilon.i,xi.i,tau.i,TRUE)
ups.c1 = chromatic1[[1]]
tau.c1 = chromatic1[[2]]
xi.c1 = chromatic1[[3]]
time.c1 = chromatic1[[4]]
G.c1 = chromatic1[[5]]
burn.c1 = chromatic1[[6]]
ups.hat.c1 = mean(ups.c1[,burn.c1:G.c1])
tau.hat.c1 = mean(tau.c1[,burn.c1:G.c1])
xi.hat.c1 = apply(xi.c1[,burn.c1:G.c1],1,mean)  


###Second Initialization
upsilon.i = log(mean(Y/mi)/(1 - mean(Y/mi)))
eta.app = log((Y/mi)/(1 - Y/mi))
ind = eta.app== -Inf
eta.app[ind] = -99
ind = eta.app== Inf
eta.app[ind] = 99
xi.vals = t(G.Mat)%*%(eta.app - upsilon.i)
xi.i = matrix(xi.vals,S,1)
tau.i = as.numeric(var(xi.i))

###Block
block2 = block.gibbs(10000,5000,upsilon.i,xi.i,tau.i,TRUE)
ups.b2 = block2[[1]]
tau.b2 = block2[[2]]
xi.b2 = block2[[3]]
time.b2 = block2[[4]]
G.b2 = block2[[5]]
burn.b2 = block2[[6]]
ups.hat.b2 = mean(ups.b2[,burn.b2:G.b2])
tau.hat.b2 = mean(tau.b2[,burn.b2:G.b2])
xi.hat.b2 = apply(xi.b2[,burn.b2:G.b2],1,mean)  

###Chromatic
chromatic2 = chromatic.gibbs(10000,5000,upsilon.i,xi.i,tau.i,TRUE)
ups.c2 = chromatic2[[1]]
tau.c2 = chromatic2[[2]]
xi.c2 = chromatic2[[3]]
time.c2 = chromatic2[[4]]
G.c2 = chromatic2[[5]]
burn.c2 = chromatic2[[6]]
ups.hat.c2 = mean(ups.c2[,burn.c2:G.c2])
tau.hat.c2 = mean(tau.c2[,burn.c2:G.c2])
xi.hat.c2 = apply(xi.c2[,burn.c2:G.c2],1,mean) 

###Third Initialization
upsilon.i = 0
eta.app = log((Y/mi)/(1 - Y/mi))
ind = eta.app== -Inf
eta.app[ind] = -99
ind = eta.app== Inf
eta.app[ind] = 99
xi.vals = t(G.Mat)%*%(eta.app)
xi.i = matrix(xi.vals,S,1)
tau.i = as.numeric(var(xi.i))

###Block
block3 = block.gibbs(10000,5000,upsilon.i,xi.i,tau.i,TRUE)
ups.b3 = block3[[1]]
tau.b3 = block3[[2]]
xi.b3 = block3[[3]]
time.b3 = block3[[4]]
G.b3 = block3[[5]]
burn.b3 = block3[[6]]
ups.hat.b3 = mean(ups.b3[,burn.b3:G.b3])
tau.hat.b3 = mean(tau.b3[,burn.b3:G.b3])
xi.hat.b3 = apply(xi.b3[,burn.b3:G.b3],1,mean)  

###Chromatic
chromatic3 = chromatic.gibbs(10000,5000,upsilon.i,xi.i,tau.i,TRUE)
ups.c3 = chromatic3[[1]]
tau.c3 = chromatic3[[2]]
xi.c3 = chromatic3[[3]]
time.c3 = chromatic3[[4]]
G.c3 = chromatic3[[5]]
burn.c3 = chromatic3[[6]]
ups.hat.c3 = mean(ups.c3[,burn.c3:G.c3])
tau.hat.c3 = mean(tau.c3[,burn.c3:G.c3])
xi.hat.c3 = apply(xi.c3[,burn.c3:G.c3],1,mean) 

